import numpy as np
from tqdm import tqdm
import tensorflow as tf
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.layers import Dense, Conv1D, BatchNormalization, Flatten
from tensorflow.keras.models import Sequential
from tensorflow.keras.initializers import GlorotNormal

class NeuralNetwork():
    """Online changepoint detection using neural networks.
    
    Read more in
    
    Parameters
    ----------
    
    k : int, default=10
        First windowing parameter which stacks d-dimensional observations in arrays
        of shape (k, d) called autoregressive forms.
    
    n : int, default=5
        Second windowing parameter which stacks autoregressive observations of dimension
        (k, d) into tensors of shape (n, k, d). Those tensors are called mini-batches.
    
    lag : int, default=100
        Lag parameter controlling the time delay between a recent and a lagged mini-batch of
        data to be compared by the neural network. Recall that one should use n << lag.
    
    f : tf.keras.models.Sequential, default=None
        Neural network architecture to learn the mini-batches distribution sequentially. Must
        be a Keras object, with input size (n, k, d) and output size (1,). The default neural
        network architecture is a 1D Convolutional neural network with 2 layers, along with 
        BatchNormalisation layers. 
    
    r : float, default=0.1
        Learning rate of the modified EWMA algorithm. 
    
    L : float, default=3.
        Bandwith parameter of the modified EWMA algorithm.
    
    burnin : int, default=100
        Burnin period during which the first observations can not contain any changepoint.
    
    method : str, default='bump'
        The modified EWMA algorithm to use to detect changepoints in the dissimilarity sequence.
        Use 'bump' to detect sequential increases and decreases, use 'increase' to detect each bump
        increases.
    
    timeout : int, default=100
        Timeout period to be used if method='increase'. When a changepoint is detected, no change
        can be detected in the following timeout data observations.
    
    Attributes
    ----------
    
    divergence : ndarray
        Kullback-Leibler divergence sequence computed at each iteration and mini-batches comparison.
    
    dissimilarity : ndarray
        Rolling Kullback-Leibler divergence on the last samples to create a dissimilarity sequence 
        containing bumps.
        
    optimiser : tf.keras.optimizers, default=tf.keras.optimizers.Adam()
        Optimiser object used for the neural network training.
        
    Z : ndarray
        Modified EWMA algorithm exponentially weighted moving average statistic.
    
    sigma_Z : ndarray
        Modified EWMA algorithm standard deviation of the Z statistic.
    
    changepoints : ndarray
        Array containing detected changepoints, initialised empty.
    """
    
    def __init__(self, 
                 k: int = 10, 
                 n: int = 5, 
                 lag:  int = 100, 
                 f: tf.keras.models.Sequential = None, 
                 r: float = 0.1, 
                 L: float = 3., 
                 burnin: int = 100, 
                 method: str = "bump", 
                 timeout: int = 100):
        self.k = k
        self.n = n
        self.l = lag
        if f is None:
            # Neural network architecture by default
            self.f = Sequential([
                Conv1D(16, 2, activation="elu", input_shape=(self.n, self.k), kernel_initializer=GlorotNormal()),
                BatchNormalization(),
                Conv1D(8, 2, activation="elu", kernel_initializer=GlorotNormal()),
                BatchNormalization(),
                Flatten(),
                Dense(1, activation="sigmoid", kernel_initializer=GlorotNormal())
            ])
        else:
            self.f = f
        self.r = r
        self.L = L
        self.burnin = burnin
        self.timeout = timeout
        if method == "bump" or method == "increase":
            self.method = method
        else:
            print("Method can either be 'bump' or 'increase'.")
            self.method = "bump"
        
        self.divergence = []
        self.dissimilarity = []
        self.optimiser = Adam()
        self.Z = [0.]
        self.sigma_Z = [0.]
        self.changepoints = []
    
    def window_data(self, 
                    data: list):
        """Window the d-dimensional stream.

        Parameters
        ----------
        
        data : list
            d-dimensional data stream to be processed to create the mini-batches of data.
        """
        T = len(data)
        self.X = []
        for i in range(T - self.k + 1):
            self.X.append(data[i:self.k+i])
        self.X = np.asarray(self.X)
        self.Xi = []
        for i in range(T - self.k - self.n + 2):
            self.Xi.append(self.X[i:self.n+i])
        self.Xi = np.asarray(self.Xi)

    def compute_dissimilarity(self):
        """Compute the dissimilarity sequence.
        """
        for i in tqdm(range(len(self.Xi) - self.l - 1)):
            X_lagged = self.Xi[i][tf.newaxis, ...]
            X_recent = self.Xi[i+self.l][tf.newaxis, ...]
            d = tf.math.log((1 - self.f(X_lagged)) / self.f(X_lagged)) + tf.math.log(self.f(X_recent) / (1 - self.f(X_recent)))
            if i <= self.l + 1:
                self.dissimilarity.append(0.)
            else:
                d_bar = self.dissimilarity[-1] + (d - self.divergence[i-1-self.l]) / self.l
                self.dissimilarity.append(d_bar.numpy()[0][0])
            self.divergence.append(d.numpy()[0][0])
            with tf.GradientTape() as tape:
                loss_value = - tf.math.log(1 - self.f(X_lagged)) - tf.math.log(self.f(X_recent))
            grads = tape.gradient(loss_value, self.f.trainable_weights)
            self.optimiser.apply_gradients(zip(grads, self.f.trainable_weights))
        self.divergence = np.asarray(self.divergence)
        self.dissimilarity = np.asarray(self.dissimilarity)
        
    def detect_increase(self):
        """Detect increases in the dissimilarity.
        """
        n = 2
        mu_hat = 0.
        sigma_hat = 1.
        last_cp = 0
        for j in range(1, self.burnin):
            self.Z.append((1 - self.r) * self.Z[j-1] + self.r * self.dissimilarity[j])
            mu_hat_new = 1/n * ((n-1) * mu_hat + self.dissimilarity[j])
            sigma_hat = np.sqrt(1 / (n - 1) * ((self.dissimilarity[j] - mu_hat_new) * 
                                             (self.dissimilarity[j] - mu_hat) + (n - 2) * sigma_hat))
            mu_hat = mu_hat_new
            n += 1
            self.sigma_Z.append(sigma_hat * np.sqrt((self.r / (2 - self.r)) * (1 - (1 - self.r) ** (2 * j))))
        for j in range(self.burnin, len(self.dissimilarity)):
            self.Z.append((1 - self.r) * self.Z[j-1] + self.r * self.dissimilarity[j])
            mu_hat_new = 1/n * ((n-1) * mu_hat + self.dissimilarity[j])
            sigma_hat = np.sqrt(1 / (n - 1) * ((self.dissimilarity[j] - mu_hat_new) * 
                                             (self.dissimilarity[j] - mu_hat) + (n - 2) * sigma_hat))
            mu_hat = mu_hat_new
            n += 1
            self.sigma_Z.append(sigma_hat * np.sqrt((self.r / (2 - self.r)) * (1 - (1 - self.r) ** (2 * j))))
            if self.Z[j] > mu_hat + self.sigma_Z[j] * self.L:
                if j - last_cp > self.timeout:
                    self.changepoints.append(j)
                    last_cp = j
                n = 2
    
    def detect_bump(self):
        """Detect sequential increase and decrease in the dissimilarity.
        """
        n = 2
        mu_hat = 0.
        sigma_hat = 1.
        last = None
        last_det = None
        for j in range(1, self.burnin):
            self.Z.append((1 - self.r) * self.Z[j-1] + self.r * self.dissimilarity[j])
            mu_hat_new = 1/n * ((n-1) * mu_hat + self.dissimilarity[j])
            sigma_hat = np.sqrt(1 / (n - 1) * ((self.dissimilarity[j] - mu_hat_new) * 
                                             (self.dissimilarity[j] - mu_hat) + (n - 2) * sigma_hat))
            mu_hat = mu_hat_new
            n += 1
            self.sigma_Z.append(sigma_hat * np.sqrt((self.r / (2 - self.r)) * (1 - (1 - self.r) ** (2 * j))))
        for j in range(self.burnin, len(self.dissimilarity)):
            self.Z.append((1 - self.r) * self.Z[j-1] + self.r * self.dissimilarity[j])
            mu_hat_new = 1/n * ((n-1) * mu_hat + self.dissimilarity[j])
            sigma_hat = np.sqrt(1 / (n - 1) * ((self.dissimilarity[j] - mu_hat_new) * 
                                             (self.dissimilarity[j] - mu_hat) + (n - 2) * sigma_hat))
            mu_hat = mu_hat_new
            n += 1
            self.sigma_Z.append(sigma_hat * np.sqrt((self.r / (2 - self.r)) * (1 - (1 - self.r) ** (2 * j))))
            if self.Z[j] > mu_hat + self.sigma_Z[j] * self.L and last_det is None:
                last = "+"
                last_det = j
                n = 2
            if self.Z[j] < mu_hat - self.sigma_Z[j] * self.L and last == "+":
                last = "-"
                n = 2
                self.changepoints.append(last_det)
                last_det = None
        
    def process(self, 
                data: list):
        """Run the online changepoint detection using neural network algorithm.

        Parameters
        ----------
        
        data : list
            Multivariate or univariate data stream to be processed. The method sequentially
            computes the Kullback-Leibler divergence between a recent mini-batch of data and 
            a lagged mini-batch of data using the neural network, which learns the distribution.
            Changepoints are detected online by detecting bumps or increases in the outputted 
            dissimilarity measure. 
        """
        # Create mini-batches of data
        self.window_data(data)
        # Compute dissimilarity measure online from mini-batches
        self.compute_dissimilarity()
        # Detect changepoints in the dissimilarity sequence
        if self.method == "bump":
            self.detect_bump()
        elif self.method == "increase":
            self.detect_increase()
        # Translate time (detection times)
        self.changepoints = np.asarray(self.changepoints) + len(data) - len(self.dissimilarity)


class EWMA():
    """Exponentially Weighted Moving Average algorithm.
        
    Read more in 
    
    Parameters
    ----------
    
    r : float, default=0.1
        Control parameter of the EWMA algorithm monitoring the learning rate
        of the exponentially moving weighted average Z (between 0 and 1).
    
    L : float, default=2.4.
        Control parameter of the EWMA algorithm used as a threshold for the 
        decision rule and controlling the bandwith.
    
    burnin : int, default=50
        Number of firts observed values processed before a changepoint can be
        detected. 
    
    mu : float, default=0.
        Initial mean value of the stream. Recall that CUSUM assumes that 
        observations are normally distributed.
    
    sigma : float, default=1.
        Initial standard deviation of the stream.
    
    Attributes
    ----------
    
    Z : ndarray of shape (n_samples,)
        Z statistic calculated sequentially after processing each new observation, 
        smoothing the original data stream.
    
    sigma_Z : ndarray of shape (n_samples,)
        Standard deviation of the Z statistic.
    
    changepoints : ndarray
        Array containing detected changepoints, initialised empty.
    
    n : int
        Number of observations in the current run length.
    """
    
    def __init__(self, 
                 r: float = 0.1, 
                 L: float = 2.4, 
                 burnin: int = 50, 
                 mu: float = 0., 
                 sigma:float = 1.):
        self.r = r
        self.L = L
        self.burnin = burnin
        self.mu = mu
        self.sigma = sigma

        self.Z = [mu]
        self.sigma_Z = [0.]
        self.changepoints = []
        self.n = 2
        
    def update_mean_variance(self, 
                             data_new: float):
        """Update efficiently mean and variance.

        Parameters
        ----------
        
        data_new : float
            New observation in the data stream. The mean and variance are updated
            efficiently and online without storing every observed values. 
        """
        mu_new = self.mu + (data_new - self.mu) / self.n
        self.sigma = (self.sigma ** 2 + ((data_new - self.mu) * (data_new - mu_new) - self.sigma ** 2) / self.n) ** 0.5
        self.mu = mu_new
    
    def update_statistics(self, 
                          i: int, 
                          data_new: float):
        """Update the algorithm statistics Z and sigma_Z.

        Parameters
        ----------
        
        i : int
            Time index.
        
        data_new : float
            New observation in the data stream. Z and sigma_Z are updated according to
            EWMA algorithm formulas.
        """
        self.Z.append((1 - self.r) * self.Z[-1] + self.r * data_new)
        self.sigma_Z.append(self.sigma * ((self.r / (2 - self.r)) * (1 - (1 - self.r) ** (2 * i))) ** 0.5)
    
    def decision_rule(self, 
                      i: int):
        """Decide whether or not a change has occurred.

        Parameters
        ----------
            
        i : int
            Time index. The decision rule |Z - mu| / sigma_Z > L is implemented in this method.
        """
        if (i >= self.burnin) and (abs((self.Z[-1] - self.mu) / self.L) > self.sigma_Z[-1]):
            self.changepoints.append(i)
            self.n = 2
        else:
            self.n += 1
        
    def process(self, 
                data: list):
        """Run EWMA algorithm on a univariate data stream.

        Parameters
        ----------
            
        data : list
            Univariate data stream to be processed. The method sequentially first updates mean 
            and variance, then updates the Z and sigma_Z statistics and finally applies the simple
            decision rule to assert if a change has occurred.
        """
        for i in range(1, len(data)):
            self.update_mean_variance(data[i])
            self.update_statistics(i, data[i])
            self.decision_rule(i)


import numpy as np
from scipy.stats import ranksums, mood, mannwhitneyu, ks_2samp, cramervonmises_2samp

class TwoSample():
    """Two sample test for changepoint detection.
    
    Read more in
    
    Parameters
    ----------
    
    statistic : str, default="Lepage"
        Test statistic to be used by the algorithm. Use 'Mann-Whitney' for changes
        in the location, 'Mood' for changes in the scale, 'Lepage' for changes in 
        both location and scale, 'Kolmogorov-Smirnov' and 'Cramer-von-Mises' for 
        general changes in distribution.
    
    threshold : float, default=3.1
        Threshold value for the test statistic, must be suited for each statistic.
        
    Attributes
    ----------
    
    t : int
        Time index.
    
    changepoints : list
        Array containing detected changepoints, initialised empty. 
    """
    
    def __init__(self, 
                 statistic: str = "Lepage", 
                 threshold: float = 3.1):
        self.threshold = threshold
        self.statistic = statistic
        
        self.t = 2
        self.changepoints = []
        
    def fetch_statistic(self):
        # Maps statistic attribute with SciPy method
        db = {
            "Mann-Whitney": mannwhitneyu, 
            "Mood": mood, 
            "Lepage": ranksums, 
            "Kolmogorov-Smirnov": ks_2samp,
            "Cramer-von-Mises": cramervonmises_2samp
        }
        self.statistic = db[self.statistic]
    
    def process_batch(self, 
                      X: list):
        """Process a data stream until a change is detected.

        Parameters
        ----------
            
        X : list
            Univariate array to be processed.

        Returns
        -------
        
        tau : int or None
            First detected changepoint in X. If no change is detected, the method
            returns None.
        """
        Dn = []
        tau = 0
        while self.t < len(X):
            Dkn = []
            for k in range(1, self.t):
                x, y = X[tau:k], X[k:self.t]
                # Different syntax for each Scipy test
                try:
                    Dkn.append(abs(self.statistic(x, y)[0]))
                except:
                    try:
                        Dkn.append(abs(self.statistic(x, y).statistic))
                    except: # Edge cases with sample sizes 
                        Dkn.append(0)
            if max(Dkn) < self.threshold:
                Dn.append(max(Dkn))
                self.t += 1
            else:
                tau = self.t
                return tau
        return None

    def process(self, 
                data: list):
        """Run the two-sample test algorithm.

        Parameters
        ----------
            
        data : list
            Univariate data stream to be processed. The method first tries to detect
            a change in the data. If a change is detected, it will look for changes in the
            remaining sequence, etc. until no change is detected. When no change is detected
            and the whole sequence has been processed, the method stops.
        """
        self.fetch_statistic()
        cp = []
        tau = 0
        while tau is not None:
            data = data[tau:]
            tau = self.process_batch(data)
            cp.append(tau)
        self.changepoints = np.cumsum(cp[:-1])


class CUSUM():
    """Cumulative Sum algorithm.
    
    Read more in 
    
    Parameters
    ----------
    
    k : float, default=0.25
        Control parameter of the CUSUM algorithm monitoring the gap between 
        the normalised stream and the algorithm statistics. 
    
    h : float, default=8.
        Control parameter of the CUSUM algorithm used as a threshold for the 
        decision rule.
    
    burnin : int, default=50
        Number of firts observed values processed before a changepoint can be
        detected. 
    
    mu : float, default=0.
        Initial mean value of the stream. Recall that CUSUM assumes that 
        observations are normally distributed.
    
    sigma : float, default=1.
        Initial standard deviation of the stream.
    
    Attributes
    ----------
    
    S : ndarray of shape (n_samples,)
        S statistic calculated sequentially after processing each new observation
        which measures increases.
    
    T : ndarray of shape (n_samples,)
        T statistics calculated sequentially after processing each new observation
        which measures decreases.
    
    changepoints : ndarray
        Array containing detected changepoints, initialised empty.
    
    n : int
        Number of observations in the current run length.
    """
    
    def __init__(self, 
                 k: float = 0.25, 
                 h: float = 8., 
                 burnin: int = 50, 
                 mu: float = 0., 
                 sigma:float = 1.):
        self.k = k
        self.h = h
        self.burnin = burnin
        self.mu = mu
        self.sigma = sigma

        self.S = [0.]
        self.T = [0.]
        self.changepoints = []
        self.n = 2
    
    def update_mean_variance(self, 
                             data_new: float):
        """Update efficiently mean and variance.

        Parameters
        ----------
        
        data_new : float
            New observation in the data stream. The mean and variance are updated
            efficiently and online without storing every observed values. 
        """
        mu_new = self.mu + (data_new - self.mu) / self.n
        self.sigma = (self.sigma ** 2 + ((data_new - self.mu) * (data_new - mu_new) - self.sigma ** 2) / self.n) ** 0.5
        self.mu = mu_new
        
    def update_statistics(self, 
                          data_new: float):
        """Update the algorithm statistics S and T.

        Parameters
        ----------
        
        data_new : float
            New observation in the data stream. S and T are updated according to
            CUSUM algorithm formulas.
        """
        self.S.append(max(0, self.S[-1] + (data_new - self.mu) / self.sigma - self.k))
        self.T.append(max(0, self.S[-1] - (data_new - self.mu) / self.sigma - self.k))
    
    def decision_rule(self, 
                      i: int):
        """Decide whether or not a change has occurred.

        Parameters
        ----------
            
        i : int
            Time index. The decision rule S > h or T > h is implemented in this method.
        """
        if (i >= self.burnin) and ((self.S[-1] > self.h) or (self.T[-1] > self.h)):
            self.changepoints.append(i)
            self.S[-1] = 0
            self.T[-1] = 0
            self.n = 2
        else:
            self.n += 1
        
    def process(self, 
                data: list):
        """Run CUSUM algorithm on a univariate data stream.

        Parameters
        ----------
            
        data : list
            Univariate data stream to be processed. The method sequentially first updates mean 
            and variance, then updates the S and T statistics and finally applies the simple
            decision rule to assert if a change has occurred.
        """
        for i in range(1, len(data)):
            self.update_mean_variance(data[i])
            self.update_statistics(data[i])
            self.decision_rule(i)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ocpdet",
    version="0.0.1",
    author="Victor Khamesi",
    author_email="victor.khamesi21@imperial.ac.uk",
    description="A Python library for online changepoint detection, implementing well-known and recent algorithms.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vkhamesi/ocpdet",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "numpy",
        "tqdm",
        "tensorflow",
        "scipy"
    ],
    python_requires='>=3.6',
)